/*
 *  CoreAudioDataLine.cpp
 *  CoreAudioJavaSoundLib
 *
 *  Created by Klaus Jänsch on 31.01.12.
 *  Copyright 2012 __MyCompanyName__. All rights reserved.
 *
 */

#include "CoreAudioDataLine.h"
#include <CoreFoundation/CFString.h>
#include <CoreFoundation/CoreFoundation.h>
#include <CoreAudio/CoreAudioTypes.h>
#include <CoreAudio/AudioHardware.h>
//#include <HostTime.h>
#include <CoreAudio/CoreAudio.h>
//#include <CoreServices/CarbonCore/Debugging.h>
#include <CoreServices/CoreServices.h>
#include <libkern/OSAtomic.h>

static jint bufferFilled(dataline *dl){
	jint filled=dl->writePos-dl->readPos;
	if(filled>dl->bufferSize){
		filled=dl->bufferSize;
	}
	return filled;
}


static void fillWithZeros(AudioBufferList *abList,UInt32 fromByte){
	UInt32 numBufs=abList->mNumberBuffers;
	UInt32 pos=0;
	for(int bi=0;bi<numBufs;bi++){
		AudioBuffer ab=abList->mBuffers[bi];
		UInt32 abSize=ab.mDataByteSize;
		
		UInt32 abPos=fromByte-pos;
		//if(abPos<0){
			//abPos=0;
		//}
		if(abPos<abSize){
			UInt32 toClear=abSize-abPos;
			memset(&((jbyte*)ab.mData)[abPos],0,toClear);
		}
		pos+=abSize;
	}
}

OSStatus audioIoProc(AudioDeviceID inDevice, const AudioTimeStamp*inNow, const AudioBufferList*inInputData, const AudioTimeStamp*inInputTime, 
					 AudioBufferList*outOutputData, 
					 const AudioTimeStamp*inOutputTime, 
					 void*inClientData){
	dataline* dl=(dataline*)inClientData;
	dl->running=true;
	jint bufferFill=bufferFilled(dl);
	if(dl->isPlayback){
		if(outOutputData!=NULL){
			jint requested=0;
			UInt32 numOutBufs=outOutputData->mNumberBuffers;
			for(int bi=0;bi<numOutBufs;bi++){
				AudioBuffer ab =outOutputData->mBuffers[bi];
				requested+=ab.mDataByteSize;
			}
			//AudioBuffer ab =outOutputData->mBuffers[0];
			if(dl->drain && bufferFill<=0){
				dl->drained=TRUE;
				fillWithZeros(outOutputData,0);
				
			}else{
				//printf("Data in %d buffers size of buf 0 %d chs %d\n",numOutBufs,ab.mDataByteSize,ab.mNumberChannels);
				
				jint read=0;
				for(int bi=0;bi<numOutBufs;bi++){
					AudioBuffer ab=outOutputData->mBuffers[bi];
					jbyte *jd=(jbyte *)ab.mData;
					jint toRead=ab.mDataByteSize;
					if(jd!=NULL){
						/*
						 if(toRead>bufferFill){
						 printf("buffer underrun: requested %d  filled %d\n",toRead,bufferFill);
						 fflush(stdout);
						 //toRead=bufferFill;
						 //ab.mDataByteSize=toRead;
						 }
						 */
						
						if(read+toRead>bufferFill){
							toRead=bufferFill-read;
						}
						jint bufReadPos=dl->readPos%dl->bufferSize;
						
						jint toRead1=toRead;
						if(bufReadPos+toRead1 >dl->bufferSize){
							toRead1=dl->bufferSize-bufReadPos;
						}
						
						memcpy(ab.mData,(void*)&(dl->buffer[bufReadPos]), toRead1);
						
						jint toRead2=toRead-toRead1;
						if(toRead2>0){
							memcpy(&(jd[toRead1]),(void*)&(dl->buffer[0]), toRead2);
						}
						
					}
					read+=toRead;
				}
				OSAtomicAdd64Barrier(read,&(dl->readPos));
				/*
				int floatSize=ab.mDataByteSize/sizeof(Float32);
				int i;
				for(i=0;i<floatSize;i++){
					Float32 *abF=(Float32*)ab.mData;
					abF[i]=0.0;
				}
				 */
			}
			
		}
	}else{
		/*
		if(inInputTime->mSampleTime<0){
			printf("Dropping first buffer(s) containing stale audio data\n");
			// may contain stale data, drop it
		}
		 */
		if(inInputData!=NULL && inInputTime->mSampleTime>=0.0){
			//printf("nowtime %f sampletime %f writePos %d\n",inNow->mSampleTime,inInputTime->mSampleTime,dl->writePos);
			UInt32 numBufs=inInputData->mNumberBuffers;
			
			jint written=0;
			for(int bi=0;bi<numBufs;bi++){
				AudioBuffer ab =inInputData->mBuffers[bi];
				
				//printf("Data in %d buffers size of buf 0 %d chs %d\n",numBufs,ab.mDataByteSize,ab.mNumberChannels);
				jbyte *jd=(jbyte *)ab.mData;
				jint toWrite=ab.mDataByteSize;
				jint bufWritePos=dl->writePos%dl->bufferSize;
				jint toWrite1=toWrite;
				
				if(bufWritePos+toWrite1 >dl->bufferSize){
					toWrite1=dl->bufferSize-bufWritePos;
					
				}
				
				memcpy((void*)&(dl->buffer[bufWritePos]),ab.mData, toWrite1);
				//memcpy((void*)&(dl->testBuffer[bufWritePos]),ab.mData, toWrite1);
				jint toWrite2=toWrite-toWrite1;
			
				if(toWrite2>0){
					memcpy((void*)&(dl->buffer[0]),&(jd[toWrite1]), toWrite2);
					//fprintf(stdout,"toWrite2 %d\n",toWrite2);
					//fflush(stdout);
				}
				written+=toWrite;
			}
			
			OSAtomicAdd64Barrier(written,&(dl->writePos));
		}
	}
	return 0;
}

static OSStatus propertyUpdate(
							   AudioDeviceID inDevice, 
							   UInt32 inChannel, 
							   Boolean isInput, 
							   AudioDevicePropertyID inPropertyID, 
							   void*inClientData){
	printf("Update\n");
	fflush(stdout);
	return 0;
}


JNIEXPORT void JNICALL Java_ips_audio_coreaudio_CoreAudioDataLine_enumerateFormats(JNIEnv *env, jobject obj,jobject jdl){
	
	dataline* dl=(dataline*)env->GetDirectBufferAddress(jdl);
	AudioStreamID streamId=dl->streamID;
	UInt32 dataSize;
	
	//AudioStreamGetPropertyInfo(streamId, 0, kAudioStreamPropertyAvailablePhysicalFormats, &dataSize, NULL);
    AudioObjectPropertyAddress pAdr;
    pAdr.mSelector=kAudioStreamPropertyAvailablePhysicalFormats;
    pAdr.mScope=kAudioObjectPropertyScopeGlobal;
    pAdr.mElement=kAudioObjectPropertyElementMaster;
    
    OSStatus res=AudioObjectGetPropertyDataSize(streamId, &pAdr, 0, NULL, &dataSize);
    if(res!=0){
        fprintf(stderr,"Could not get stream property data size.");
        fflush(stderr);
		return;
    }
	AudioStreamRangedDescription* rds=(AudioStreamRangedDescription*)malloc(dataSize);
	//AudioStreamGetProperty(streamId,0,kAudioStreamPropertyAvailablePhysicalFormats,&dataSize,rds);
    AudioObjectGetPropertyData(streamId, &pAdr, 0, NULL, &dataSize,rds);
	jclass cls=env->FindClass("ips/audio/coreaudio/CoreAudioDataLine");
	if(cls==NULL){
		fprintf(stderr,"Java Class not found!");
		return;
	}
	jmethodID mId=env->GetMethodID(cls,"nextFormat","(FII)V");
	if(mId==NULL){
		fprintf(stderr,"Method not found!");
		return;
	}
	jmethodID mIdr=env->GetMethodID(cls,"nextFormatRange","(FFII)V");
	if(mIdr==NULL){
		fprintf(stderr,"Method not found!");
		return;
	}
	for(int j=0;j<dataSize/(sizeof(AudioStreamRangedDescription));j++){
		AudioStreamRangedDescription rd=rds[j];
		AudioStreamBasicDescription af=rd.mFormat;
		Float64 sampleRate=af.mSampleRate;
		if(sampleRate==kAudioStreamAnyRate){
			AudioValueRange sampleRateRange=rd.mSampleRateRange;
			env->CallVoidMethod(obj,mIdr,sampleRateRange.mMinimum,sampleRateRange.mMaximum,af.mBitsPerChannel,af.mChannelsPerFrame);
		}else{
		//printf("RD: %d %f bytesPerFrame %d bytesPPacket %d flags %d framePPacket %d\n",af.mChannelsPerFrame,af.mSampleRate,af.mBytesPerFrame,af.mBytesPerPacket,af.mFormatFlags,af.mFramesPerPacket);
			env->CallVoidMethod(obj,mId,af.mSampleRate,af.mBitsPerChannel,af.mChannelsPerFrame);
		}
	}
	
	
}

static jobject dataLineOpen(JNIEnv *env, jobject obj,jobject jdl,jfloat sampleRate,jint sampleSizeInBits,jint channels,jint bufferSize){
	
	dataline* dl=(dataline*)env->GetDirectBufferAddress(jdl);
	dl->sampleSizeInBytes=sampleSizeInBits/8;
	jint frameSize=dl->sampleSizeInBytes*channels;
	dl->pFormat.mFormatFlags = kAudioFormatFlagIsSignedInteger | kAudioFormatFlagIsPacked;
	//dl->pFormat.mFormatFlags = 12;
	dl->pFormat.mFormatID=kAudioFormatLinearPCM;
	dl->pFormat.mSampleRate=sampleRate;
	dl->pFormat.mBytesPerPacket=frameSize;
	dl->pFormat.mFramesPerPacket=1;
	//dl->pFormat.mFramesPerPacket=12;
	dl->pFormat.mBytesPerFrame=frameSize;
	dl->pFormat.mChannelsPerFrame=channels;
	dl->pFormat.mBitsPerChannel=sampleSizeInBits;
	//printf("Audio format %f bits %d frameSize %d channels %d\n",sampleRate,sampleSizeInBits,frameSize,channels);
	
	AudioStreamID streamId=dl->streamID;
	dl->bufferSize=bufferSize;
	dl->buffer=(jbyte *)malloc(bufferSize);
	memset(dl->buffer,0,bufferSize);
	dl->writePos=0;
	dl->readPos=0;
	dl->drain=FALSE;
	dl->drained=FALSE;
	
    AudioObjectPropertyAddress pAdr;
    pAdr.mSelector=kAudioStreamPropertyDirection;
    pAdr.mScope=kAudioObjectPropertyScopeGlobal;
    pAdr.mElement=kAudioObjectPropertyElementMaster;
   	AudioStreamBasicDescription currFormat;
	UInt32 propSize=sizeof(AudioStreamBasicDescription);
    pAdr.mSelector=kAudioStreamPropertyPhysicalFormat;
	
    OSStatus res=AudioObjectGetPropertyData(streamId, &pAdr, 0, NULL, &propSize, &currFormat);
	if(res!=0){
        fprintf(stderr,"Could not get physical audio format.\n");
        fflush(stderr);
        return NULL;
	}
	
	// Set the physical format only if it has hacnged to avoid switch-on noise on start of playback 
	// compare format
	AudioStreamBasicDescription reqFmt=dl->pFormat;
	
	if(res!=0 || reqFmt.mFormatFlags!=currFormat.mFormatFlags ||
	   reqFmt.mFormatID!=currFormat.mFormatID ||
	   reqFmt.mSampleRate!=currFormat.mSampleRate ||
	   reqFmt.mBytesPerPacket!=currFormat.mBytesPerPacket ||
	   reqFmt.mFramesPerPacket != currFormat.mFramesPerPacket ||
	   reqFmt.mBytesPerFrame != currFormat.mBytesPerFrame ||
	   reqFmt.mChannelsPerFrame != currFormat.mChannelsPerFrame ||
	   reqFmt.mBitsPerChannel!= currFormat.mBitsPerChannel){
        pAdr.mSelector=kAudioStreamPropertyPhysicalFormat;
		//res=AudioStreamSetProperty(streamId, NULL,0,kAudioStreamPropertyPhysicalFormat, sizeof ( AudioStreamBasicDescription ),&(dl->pFormat));
        res=AudioObjectSetPropertyData(streamId, &pAdr, 0, NULL, sizeof(AudioStreamBasicDescription), &(dl->pFormat));
		if(res!=0){
			//const char * errStr=GetMacOSStatusErrorString(res);
			fprintf(stderr,"Could not set physical audio format\n");
            fflush(stderr);
			return NULL;
		}
	}
	res = AudioDeviceCreateIOProcID(dl->deviceID, audioIoProc, dl,&(dl->ioProcId));
	if(res!=0){
        fprintf(stderr,"Could not get audio io proc ID\n");
        fflush(stderr);
		return NULL;
	}

	jobject jrb=env->NewDirectByteBuffer(dl->buffer,bufferSize);
	return jrb;
}
JNIEXPORT jobject JNICALL Java_ips_audio_coreaudio_CoreAudioDataLine_open(JNIEnv *env, jobject obj,jobject jdl,jfloat sampleRate,jint sampleSizeInBits,jint channels,jint bufferSize){
	return dataLineOpen(env,obj,jdl,sampleRate,sampleSizeInBits,channels,bufferSize);
}

JNIEXPORT jlong JNICALL Java_ips_audio_coreaudio_CoreAudioDataLine_getWritePos(JNIEnv *env, jobject obj,jobject jdl){
	dataline* dl=(dataline*)env->GetDirectBufferAddress(jdl);
	return dl->writePos;
	
}

JNIEXPORT jlong JNICALL Java_ips_audio_coreaudio_CoreAudioDataLine_getReadPos(JNIEnv *env, jobject obj,jobject jdl){
	dataline* dl=(dataline*)env->GetDirectBufferAddress(jdl);
	return dl->readPos;
	
}

JNIEXPORT void JNICALL Java_ips_audio_coreaudio_CoreAudioTargetDataLine_addToReadPos(JNIEnv *env, jobject obj,jobject jdl,jint read){
	dataline* dl=(dataline*)env->GetDirectBufferAddress(jdl);
	OSAtomicAdd64Barrier(read,&(dl->readPos));
	
}
JNIEXPORT void JNICALL Java_ips_audio_coreaudio_CoreAudioSourceDataLine_addToWritePos(JNIEnv *env, jobject obj,jobject jdl,jint written){
	dataline* dl=(dataline*)env->GetDirectBufferAddress(jdl);
	OSAtomicAdd64Barrier(written,&(dl->writePos));
	
}


JNIEXPORT void JNICALL Java_ips_audio_coreaudio_CoreAudioTargetDataLine_flush(JNIEnv *env, jobject obj,jobject jdl){
	dataline* dl=(dataline*)env->GetDirectBufferAddress(jdl);
	// not full snchronized
	dl->readPos=dl->writePos;
}

JNIEXPORT void JNICALL Java_ips_audio_coreaudio_CoreAudioSourceDataLine_flush(JNIEnv *env, jobject obj,jobject jdl){
	dataline* dl=(dataline*)env->GetDirectBufferAddress(jdl);
	// not full snchronized
	dl->writePos=dl->readPos;
}

JNIEXPORT jboolean JNICALL Java_ips_audio_coreaudio_CoreAudioDataLine_start(JNIEnv *env, jobject obj,jobject jdl){
	dataline* dl=(dataline*)env->GetDirectBufferAddress(jdl);
	AudioDeviceID devId=dl->deviceID;
	OSStatus	st=AudioDeviceStart(devId,audioIoProc);
    if(st != 0){
		printf("Error on CoreAudio start\n");
	}
	return (st==0);	
}

JNIEXPORT jboolean JNICALL Java_ips_audio_coreaudio_CoreAudioTargetDataLine_isRunning(JNIEnv *env, jobject obj,jobject jdl){
	dataline* dl=(dataline*)env->GetDirectBufferAddress(jdl);
	return dl->running;
}

JNIEXPORT jint JNICALL Java_ips_audio_coreaudio_CoreAudioTargetDataLine_coreAudioBytesAvailable(JNIEnv *env, jobject obj,jobject jdl){
	dataline* dl=(dataline*)env->GetDirectBufferAddress(jdl);
	return bufferFilled(dl);
}

JNIEXPORT jint JNICALL Java_ips_audio_coreaudio_CoreAudioSourceDataLine_coreAudioBytesAvailable(JNIEnv *env, jobject obj,jobject jdl){
	dataline* dl=(dataline*)env->GetDirectBufferAddress(jdl);
	jint filled=bufferFilled(dl);
	return dl->bufferSize-filled;
}

JNIEXPORT void JNICALL Java_ips_audio_coreaudio_CoreAudioSourceDataLine_drain(JNIEnv *env, jobject obj,jobject jdl){
	dataline* dl=(dataline*)env->GetDirectBufferAddress(jdl);
	dl->drain=TRUE;
}

JNIEXPORT jboolean JNICALL Java_ips_audio_coreaudio_CoreAudioSourceDataLine_drained(JNIEnv *env, jobject obj,jobject jdl){
	dataline* dl=(dataline*)env->GetDirectBufferAddress(jdl);
	return dl->drained;
}

JNIEXPORT jboolean JNICALL Java_ips_audio_coreaudio_CoreAudioDataLine_stop(JNIEnv *env, jobject obj,jobject jdl){
	dataline* dl=(dataline*)env->GetDirectBufferAddress(jdl);
	AudioDeviceID devId=dl->deviceID;
	OSStatus st=AudioDeviceStop(devId,dl->ioProcId);
    if(st != 0){
		fprintf(stderr,"Error on CoreAudio stop\n");
	}
    return (st==0);
	
}

JNIEXPORT void JNICALL Java_ips_audio_coreaudio_CoreAudioDataLine_close(JNIEnv *env, jobject obj,jobject jdl){
	dataline* dl=(dataline*)env->GetDirectBufferAddress(jdl);
	AudioDeviceID devId=dl->deviceID;
	AudioDeviceDestroyIOProcID(devId, dl->ioProcId);
	dl->running=false;
	free(dl->buffer);
	
	
}


